import psutil


def get_available_cpu_memory():
    this_process = psutil.Process()
    available_memory = psutil.virtual_memory().available

    try:
        import resource

        soft_mem_limit, hard_mem_limit = resource.getrlimit(resource.RLIMIT_AS)
        if hard_mem_limit != resource.RLIM_INFINITY:
            used_memory = this_process.memory_info().vms
            available_memory = min(hard_mem_limit - used_memory, available_memory)
    except ImportError:
        pass

    return available_memory


def set_cpu_memory_limit(num_gigabytes):
    try:
        import resource

        num_bytes = int(num_gigabytes * 2 ** 30)
        _, hard_limit = resource.getrlimit(resource.RLIMIT_AS)
        if hard_limit != resource.RLIM_INFINITY:
            hard_limit = min(num_bytes, hard_limit)
        else:
            hard_limit = num_bytes
        resource.setrlimit(resource.RLIMIT_AS, (hard_limit, hard_limit))
    except ImportError:
        pass


def is_out_of_cpu_memory(exception):
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "DefaultCPUAllocator: can't allocate memory" in exception.args[0]
    )



"""Torch Memory-Adaptive Algorithms.

Helpers to allow for OOM conditions and dynamic adaptation of "internal"
batchsizes. (Without affecting the computational ones.)
"""
import functools
from typing import Type, Optional

import torch

import toma.stacktrace as tst
from toma.batchsize_cache import StacktraceMemoryBatchsizeCache, NoBatchsizeCache, GlobalBatchsizeCache
from toma.cpu_memory import is_out_of_cpu_memory
from toma.torch_cuda_memory import is_cuda_out_of_memory, is_cudnn_snafu, gc_cuda


DEFAULT_CACHE_TYPE = StacktraceMemoryBatchsizeCache


class simple:
    """
    Straight-forward wrappers (can be copy-pasted and hacked easily).
    """

    @staticmethod
    def batch(func, initial_batchsize: int, *args, **kwargs):
        gc_cuda()

        batchsize = initial_batchsize
        while True:
            try:
                return func(batchsize, *args, **kwargs)
            except RuntimeError as exception:
                if batchsize > 1 and should_reduce_batch_size(exception):
                    batchsize //= 2
                    gc_cuda()
                else:
                    raise

    @staticmethod
    def range(func, start: int, end: int, initial_step: int, *args, **kwargs):
        gc_cuda()

        stepsize = initial_step
        current = start
        while current < end:
            try:
                func(current, min(current + stepsize, end), *args, **kwargs)
                current += stepsize
            except RuntimeError as exception:
                if stepsize > 1 and should_reduce_batch_size(exception):
                    stepsize //= 2
                    gc_cuda()
                else:
                    raise

    @staticmethod
    def chunked(func, tensor, initial_step: int, dimension: int = 0):
        def body(start, end):
            return func(tensor.narrow(dim=dimension, start=start, length=end - start), start, end)

        return simple.range(body, 0, tensor.shape[dimension], initial_step)


class toma:
    """
    Decorators that make it easy to wrap functions.
    """

    @staticmethod
    def batch(func=None, *, initial_batchsize=None, cache_type=DEFAULT_CACHE_TYPE):
        if func is None:
            return functools.partial(toma.batch, initial_batchsize=initial_batchsize, cache_type=cache_type)

        @functools.wraps(func)
        def wrapped(*args, toma_initial_batchsize=None, **kwargs):
            _initial_batchsize = toma_initial_batchsize or initial_batchsize
            return explicit.batch(func, _initial_batchsize, *args, toma_cache_type=cache_type, **kwargs)

        wrapped.__doc__ = f"""
Wrapped in toma.batch: 

Additional keyargs:
    toma_initial_batchsize: initial step size to use.

{wrapped.__doc__}
"""

        return wrapped

    @staticmethod
    def range(func=None, *, initial_step: Optional[int] = None, cache_type=DEFAULT_CACHE_TYPE):
        if func is None:
            return functools.partial(toma.range, initial_step=initial_step, cache_type=cache_type)

        @functools.wraps(func)
        def wrapped(start: int, end: int, *args, toma_initial_step: Optional[int] = None, **kwargs):
            _initial_step = toma_initial_step or initial_step

            return explicit.range(
                func, start, end, _initial_step, *args, toma_context=func, toma_cache_type=cache_type, **kwargs
            )

        wrapped.__doc__ = f"""
Wrapped in toma.range: 

Additional keyargs:
    toma_initial_step: initial step size to use.

{wrapped.__doc__}
"""

        return wrapped

    @staticmethod
    def chunked(
        func=None,
        *,
        initial_step: Optional[int] = None,
        dimension: Optional[int] = None,
        cache_type: Type = DEFAULT_CACHE_TYPE,
    ):
        dimension = dimension or 0
        if func is None:
            return functools.partial(
                toma.chunked, initial_step=initial_step, dimension=dimension, cache_type=cache_type
            )

        @functools.wraps(func)
        def wrapped(
            tensor: torch.Tensor,
            *args,
            toma_initial_step: Optional[int] = None,
            toma_dimension: Optional[int] = None,
            **kwargs,
        ):
            _initial_step = toma_initial_step or initial_step
            _dimension = toma_dimension or dimension

            explicit.chunked(
                func, tensor, _initial_step, *args, toma_dimension=toma_dimension, toma_cache_type=cache_type, **kwargs
            )

        wrapped.__doc__ = f"""
Wrapped in toma.chunked: 

Additional keyargs:
    toma_initial_step: initial step size to use
    toma_dimension: dimension of the tensor to chunk along

{wrapped.__doc__}
"""

        return wrapped


CONTEXT_CACHE_SIZE = 2 ** 14


@functools.lru_cache(CONTEXT_CACHE_SIZE)
def get_cache_for_context(batchsize_cache_type, context):
    return batchsize_cache_type()


class explicit:
    """
    Explicit calls that can use different cache types to memorize settings.
    """

    @staticmethod
    def batch(
        func, initial_batchsize: int, *args, toma_context=None, toma_cache_type: Type = DEFAULT_CACHE_TYPE, **kwargs
    ):
        gc_cuda()

        cache = get_cache_for_context(toma_cache_type, toma_context or func)

        batchsize = cache.get_batchsize(initial_batchsize)

        while True:
            try:
                value = batchsize.get()
                return func(value, *args, **kwargs)
            except RuntimeError as exception:
                if value > 1 and should_reduce_batch_size(exception):
                    batchsize.decrease_batchsize()
                    gc_cuda()
                else:
                    raise

    @staticmethod
    def range(
        func,
        start: int,
        end: int,
        initial_step: int,
        *args,
        toma_context=None,
        toma_cache_type: Type = DEFAULT_CACHE_TYPE,
        **kwargs,
    ):
        gc_cuda()

        cache = get_cache_for_context(toma_cache_type, toma_context or func)

        batchsize = cache.get_batchsize(initial_step)

        gc_cuda()
        current = start
        while current < end:
            try:
                func(current, min(current + batchsize.get(), end), *args, **kwargs)
                current += batchsize.get()
            except RuntimeError as exception:
                if batchsize.get() > 1 and should_reduce_batch_size(exception):
                    batchsize.decrease_batchsize()
                    gc_cuda()
                else:
                    raise

    @staticmethod
    def chunked(
        func,
        tensor: torch.Tensor,
        initial_step: int,
        *args,
        toma_dimension: int = None,
        toma_context=None,
        toma_cache_type: Type = DEFAULT_CACHE_TYPE,
        **kwargs,
    ):
        toma_dimension = toma_dimension or 0

        def body(start: int, end: int):
            return func(tensor.narrow(dim=toma_dimension, start=start, length=end - start), start, end, *args, **kwargs)

        explicit.range(
            body,
            0,
            tensor.shape[toma_dimension],
            initial_step,
            *args,
            toma_context=toma_context or func,
            toma_cache_type=toma_cache_type,
        )


def should_reduce_batch_size(exception):
    return is_cuda_out_of_memory(exception) or is_cudnn_snafu(exception) or is_out_of_cpu_memory(exception)



import functools
import inspect
from contextlib import contextmanager

__watermark = 0


def _constant_code_context(code_context):
    if not code_context:
        return None
    if len(code_context) == 1:
        return code_context[0]
    return tuple(code_context)


def get_simple_traceback(ignore_top=0):
    """Get a simple trackback that can be hashed and won't create reference
    cyles."""
    stack = inspect.stack(context=1)[ignore_top + 1 : -__watermark - 1]
    simple_traceback = tuple(
        (fi.filename, fi.lineno, fi.function, _constant_code_context(fi.code_context), fi.index) for fi in stack
    )
    return simple_traceback


@contextmanager
def watermark():
    global __watermark
    old_watermark = __watermark

    # Remove the entries for `watermark` and
    # `contextmanager.__enter__` and for the with block.
    # Remove another one to keep the caller.
    __watermark = len(inspect.stack(context=0)) - 4

    try:
        yield
    finally:
        __watermark = old_watermark


def set_watermark(func):
    @functools.wraps(func)
    def watermark_wrapper(*args, **kwargs):
        global __watermark
        old_watermark = __watermark

        # Remove the entries for `watermark` and
        # `contextmanager.__enter__`.
        # Dump frames for this wrapper.
        __watermark = len(inspect.stack(context=0)) - 1

        try:
            return func(*args, **kwargs)
        finally:
            __watermark = old_watermark

    return watermark_wrapper



"""
Helper to free Torch cuda memory and determine when a Torch exception might be
because of OOM conditions.
"""
from __future__ import print_function
import torch
import gc


def gc_cuda():
    """Gargage collect Torch (CUDA) memory."""
    gc.collect()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()


def get_cuda_total_memory():
    if torch.cuda.is_available():
        return torch.cuda.get_device_properties(0).total_memory
    return 0


def get_cuda_assumed_available_memory():
    if torch.cuda.is_available():
        return get_cuda_total_memory() - torch.cuda.memory_reserved()
    return 0


def get_cuda_available_memory():
    # Always allow for 1 GB overhead.
    if torch.cuda.is_available():
        return get_cuda_assumed_available_memory() - get_cuda_blocked_memory()
    return 0


def get_cuda_blocked_memory():
    if not torch.cuda.is_available():
        return 0

    # In GB steps
    available_memory = get_cuda_assumed_available_memory()
    current_block = available_memory - 2 ** 30
    while True:
        try:
            block = torch.empty((current_block,), dtype=torch.uint8, device="cuda")
            break
        except RuntimeError as exception:
            if is_cuda_out_of_memory(exception):
                current_block -= 2 ** 30
                if current_block <= 0:
                    return available_memory
            else:
                raise
    block = None
    gc_cuda()
    return available_memory - current_block


def is_cuda_out_of_memory(exception):
    return (
        isinstance(exception, RuntimeError) and len(exception.args) == 1 and "CUDA out of memory." in exception.args[0]
    )


def is_cudnn_snafu(exception):
    # For/because of https://github.com/pytorch/pytorch/issues/4107
    return (
        isinstance(exception, RuntimeError)
        and len(exception.args) == 1
        and "cuDNN error: CUDNN_STATUS_NOT_SUPPORTED." in exception.args[0]
    )


def cuda_meminfo():
    if not torch.cuda.is_available():
        return

    print(
        "Total:", torch.cuda.memory_allocated() / 2 ** 30, " GB Cached: ", torch.cuda.memory_reserved() / 2 ** 30, "GB"
    )
    print(
        "Max Total:",
        torch.cuda.max_memory_allocated() / 2 ** 30,
        " GB Max Cached: ",
        torch.cuda.max_memory_reserved() / 2 ** 30,
        "GB",
    )



# Always prefer setuptools over distutils
from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="toma",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="1.0.1",
    description="Write algorithms in PyTorch that adapt to the available (CUDA) memory",
    # Fix windows newlines.
    long_description=long_description.replace("\r\n", "\n"),
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url="https://github.com/blackhc/toma",
    # Author details
    author="Andreas @blackhc Kirsch",
    author_email="blackhc+toma@gmail.com",
    # Choose your license
    license="MIT",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
    # What does your project relate to?
    keywords="tools pytorch",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=["toma"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["torch", "psutil"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "dev": ["check-manifest"],
        "test": ["coverage", "codecov", "pytest", "pytest-benchmark", "pytest-cov"],
    },
    setup_requires=["pytest-runner"],
)



import functools
from dataclasses import dataclass
from typing import Optional

import toma.cpu_memory
from toma import stacktrace as tst, torch_cuda_memory as tcm
import weakref


@dataclass
class Batchsize:
    value: Optional[int] = None

    def set_initial_batchsize(self, initial_batchsize: int):
        if not self.value:
            self.value = initial_batchsize

    def get(self) -> int:
        return self.value

    def decrease_batchsize(self):
        self.value //= 2
        assert self.value > 0


class BatchsizeCache:
    all_instances = weakref.WeakValueDictionary()

    def __init__(self):
        stacktrace = tst.get_simple_traceback(2)
        BatchsizeCache.all_instances[stacktrace] = self

    def get_batchsize(self, initial_batchsize: int) -> Batchsize:
        raise NotImplementedError()


@dataclass
class NoBatchsizeCache(BatchsizeCache):
    def get_batchsize(self, initial_batchsize: int) -> Batchsize:
        return Batchsize(initial_batchsize)


@dataclass
class GlobalBatchsizeCache(BatchsizeCache):
    batchsize: Optional[Batchsize] = None

    def get_batchsize(self, initial_batchsize: int) -> Batchsize:
        if not self.batchsize:
            self.batchsize = Batchsize(initial_batchsize)
        return self.batchsize


class StacktraceMemoryBatchsizeCache(BatchsizeCache):
    LRU_CACHE_SIZE: int = 128
    MEMORY_GRANULARITY: int = 2 ** 28
    TRACK_RAM: bool = True

    initial_batchsize: Optional[int]

    def __init__(self, lru_cache_size=None):
        super().__init__()

        self.initial_batchsize = None

        @functools.lru_cache(lru_cache_size or StacktraceMemoryBatchsizeCache.LRU_CACHE_SIZE)
        def get_batchsize_from_cache(stacktrace, cpu_available_memory, gpu_available_memory):
            return Batchsize(self.initial_batchsize)

        self.get_batchsize_from_cache = get_batchsize_from_cache

    def get_batchsize(self, initial_batchsize: int):
        stacktrace = tst.get_simple_traceback(2)

        if self.TRACK_RAM:
            cpu_available_memory = int(toma.cpu_memory.get_available_cpu_memory() // self.MEMORY_GRANULARITY)
        else:
            cpu_available_memory = -1

        gpu_available_memory = int(tcm.get_cuda_assumed_available_memory() // self.MEMORY_GRANULARITY)

        batchsize = self.get_batchsize_from_cache(stacktrace, cpu_available_memory, gpu_available_memory)
        batchsize.set_initial_batchsize(initial_batchsize)
        return batchsize
